#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import zlib
import glob
import re
import datetime
import socket

import iecommon
import fesaTemplates
import fesaGeneral
import iefiles

from iecommon import *
from fesaGeneral import *
from model.Register import *
import libxml2

def findBlockServerSetActionName(fesaRoot, blockName):
    properties = fesaRoot.xpathEval("/equipment-model/interface/device-interface/*/*[@name='" + blockName + "']")
    for property in properties:
        return property.xpathEval("set-action/server-action-ref")[0].prop("server-action-name-ref")
    raise Exception("Error: Server Action for Block '" + blockName + "' not found")

#-------------------------------------------------------------------------  
# Generates the H source file containing general methods
# to synchronise the FESA fields and related PLC registers
# of the FESA server
#-------------------------------------------------------------------------  
def genHSource(className, silecsRoot, fesaRoot, sourcePath,logTopics):
    source =  fesaTemplates.genHTop(className)
    
    for block in silecsRoot.xpathEval('//Block'):
        if block.prop('mode') == 'WRITE-ONLY' or block.prop('mode') == 'READ-WRITE':
            serverActionName = findBlockServerSetActionName(fesaRoot,getFesaName(block))
            source += fesaTemplates.genHTopBlock(className, serverActionName)
        
    source += fesaTemplates.genHTop2(className)
        
    for block in silecsRoot.xpathEval('//Block'):
        if block.prop('mode') == 'READ-ONLY':
            source += fesaTemplates.genHBlock('RO', block.prop('name'),getFesaName(block) )
        elif block.prop('mode') == 'WRITE-ONLY':
            source += fesaTemplates.genHBlock('WO', block.prop('name'),getFesaName(block))
        else:   # READ-WRITE
            source += fesaTemplates.genHBlock('RW', block.prop('name'),getFesaName(block))
    
    source += fesaTemplates.genHBottom(className)

    for block in silecsRoot.xpathEval('//Block'):
        source += fesaTemplates.genHDeclBlocks(block.prop('name'))
    
    source += fesaTemplates.genHClosing(className)
    
    # Create output directory if necessary
    if not os.path.exists(sourcePath):
        os.makedirs(sourcePath)
    iecommon.logDebug("Create directory %s" %sourcePath, logTopics)
    
    # Write to file and save
    sourceFile = sourcePath + "/"+ className + ".h"
    iecommon.logInfo("Generate header file: " + sourceFile, logTopics)
    fdesc = open(sourceFile, "w")
    fdesc.write(source)
    fdesc.close()  
 
    iecommon.logInfo('Header file for '+className+' generated successfully', logTopics)
    
#-------------------------------------------------------------------------  
# Generates the C++ source file containing general 
# methods to synchronise the FESA fields and related PLC 
# registers of the FESA server
#-------------------------------------------------------------------------
    
def genCppSource(className, silecsRoot, fesaRoot, sourcePath,logTopics):
    finalSource =  fesaTemplates.genCTop(className)
    blockList = silecsRoot.xpathEval('//Block')
    for block in blockList:
        finalSource +=  fesaTemplates.genCGlobal(className, block.prop('name'))
    
    finalSource += fesaTemplates.genCPart1(className)
    
    for block in blockList:
        finalSource += fesaTemplates.genCBlockConstr(block.prop('name'), className)
        
    finalSource += fesaTemplates.genCPart2(className)
    
    for block in blockList:
        if (block.prop('mode') == 'WRITE-ONLY' or block.prop('mode') == 'READ-WRITE'):
            # just set the fields if block is WO or RW and register synchronisation is SLAVE
            # WARNING: In order to have multiplexed FESA fields, the corresponding SILECS registers' synchro mode must be 'NONE' 
            finalSource += fesaTemplates.genCSetPLC(className, block.prop('name'))
        
    finalSource += fesaTemplates.genCPart3(className)
    
    for block in blockList:

        if (block.prop('mode') == 'READ-ONLY' or block.prop('mode') == 'READ-WRITE'):
            # just get the fields if block is RO or RW and register synchronisation is MASTER
            # WARNING: In order to have multiplexed FESA fields, the corresponding SILECS registers' synchro mode must be 'NONE'
            finalSource += fesaTemplates.genCGetPLC(className, block.prop('name'))
    
    finalSource += fesaTemplates.genCPart4(className)
    
    for block in blockList:
        if block.prop('mode') != 'WRITE-ONLY':
            finalSource += fesaTemplates.genCCommonGet(block.prop('name'),className)
            finalSource += fesaTemplates.cRegVar
            finalSource += fesaTemplates.cGetArrayVar
            finalSource += fesaTemplates.cGetArray2DVar
            finalSource += '\n'
            finalSource += fesaTemplates.cRecv
            for registerNode in  block.xpathEval('Register'):
                register = DesignRegister(registerNode)
                if register.valueType == 'string':
                    finalSource += fesaTemplates.genCGetStringReg(register)
                elif register.valueType == 'stringArray':
                    finalSource += fesaTemplates.genCGetStringArrayReg(register)
                elif register.valueType == 'stringArray2D':
                    iecommon.logError('ERROR: In register '+register.name+' - 2D array of strings not supported in FESA.', True, {'errorlog': True})
                elif register.valueType == 'scalar':
                    finalSource += fesaTemplates.genCGetScalarReg(register)
                elif register.valueType == 'array':
                    finalSource += fesaTemplates.genCGetArrayReg(register)
                elif register.valueType == 'array2D':
                    finalSource += fesaTemplates.genCGetArray2DReg(register)
            finalSource += '\n    }\n'

        if block.prop('mode') != 'READ-ONLY':
            finalSource += fesaTemplates.genCCommonSet(block.prop('name'),className)
            finalSource += fesaTemplates.cRegVar
            finalSource += fesaTemplates.cSetArrayVar
            finalSource += fesaTemplates.cSetArray2DVar
            for registerNode in  block.xpathEval('Register'):
                register = DesignRegister(registerNode)
                if register.valueType == 'string':
                    finalSource += fesaTemplates.genCSetStringReg(register)
                elif register.valueType == 'stringArray':
                    finalSource += fesaTemplates.genCSetStringArrayReg(register)
                elif register.valueType == 'stringArray2D':
                    iecommon.logError('ERROR: In register '+register.name+' - 2D array of strings not supported in FESA.', True, {'errorlog': True})
                elif register.valueType == 'scalar':
                    finalSource += fesaTemplates.genCSetScalarReg(register)
                elif register.valueType == 'array':
                    finalSource += fesaTemplates.genCSetArrayReg(register)
                elif register.valueType == 'array2D':
                    finalSource += fesaTemplates.genCSetArray2DReg(register)
            finalSource += fesaTemplates.cSend
            finalSource += '    }\n'
            finalSource += fesaTemplates.genCDatatypeSet(block.prop('name'),getFesaName(block), className)
            finalSource += fesaTemplates.cRegVar
            finalSource += fesaTemplates.cSetArrayVar
            finalSource += fesaTemplates.cSetArray2DVar
            
            for registerNode in block.xpathEval('Register'):
                register = DesignRegister(registerNode)
                if register.valueType == 'string':
                    finalSource += fesaTemplates.genCSetStringRegData(register)
                elif register.valueType == 'stringArray':
                    finalSource += fesaTemplates.genCSetStringArrayRegData(register)
                elif register.valueType == 'stringArray2D':
                    iecommon.logError('ERROR: In register '+register.name+' - 2D array of strings not supported in FESA.', True, {'errorlog': True})
                elif register.valueType == 'array2D':
                    finalSource += fesaTemplates.genCSetArray2DRegData(register)
                elif register.valueType == 'array':
                    finalSource += fesaTemplates.genCSetArrayRegData(register)
                elif register.valueType == 'scalar':
                    finalSource += fesaTemplates.genCSetScalarRegData(register)

            finalSource += fesaTemplates.cSend
            finalSource += '\n    }\n'   # closing bracket for block

    finalSource += '\n}\n'   # closing bracket for class

    # Write to file and save
    sourceFile = sourcePath + "/" + className + ".cpp"
    iecommon.logInfo("Generate source file: " + sourceFile, logTopics)
    fdesc = open(sourceFile, "w")
    fdesc.write(finalSource)
    fdesc.close()

    iecommon.logInfo('Source file for '+className+' generated successfully', logTopics)

def genCppFiles(className, workspacePath, silecsDesignFilePath,logTopics={'errorlog': True}):
    fesaCommonDirectory = iefiles.getFesa3CommonDirectory(workspacePath,className)
    iecommon.logInfo("fesaCommonDirectory:" + fesaCommonDirectory,logTopics)

    if not os.path.exists(fesaCommonDirectory ):
        os.makedirs(fesaCommonDirectory)

    silecsDesignFilePath = iefiles.getSilecsDesignFilePath(workspacePath,className)
    fesaDesignFilePath = workspacePath + "/" + className + "/src/" + className + ".design"
    silecsRoot = libxml2.parseFile(silecsDesignFilePath)
    fesaRoot = libxml2.parseFile(fesaDesignFilePath)

    genHSource(className, silecsRoot, fesaRoot, fesaCommonDirectory,logTopics)
    genCppSource(className, silecsRoot, fesaRoot, fesaCommonDirectory,logTopics)

