#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import zlib
import glob
import datetime
import shutil
import libxml2

import iecommon
import xmltemplate
import iefiles

from iecommon import *
from model.Register import *

#-------------------------------------------------------------------------
# Global definitions
#-------------------------------------------------------------------------
msize                   = 0 # my regMemSize in the former PERL script
blkMemSize              = 0 # my $blockMemSize = 0; in the former PERL script
blkAddr                 = 0 # my $blockAddress = 0; in the former PERL script
deviceMemSize           = 0 # global memory-size of one class instance (sum of block-memory size)
nbBlock                 = 0 # number of block of the class
plcSize                 = 0
plcLast                 = 0
classMem                = 0
instAddr                = 0 

blockCounter            = 0 # used for NI block address generation
regCounter              = 0 # used for NI register address generation


#=========================================================================
# Sub-function
#=========================================================================

#-------------------------------------------------------------------------
# Decode HTML mapping < , > , " , ' in the appropiate coding
def decode(html):
    html = html.replace("<","&lt;")
    html = html.replace(">","&gt;")
    html = html.replace('"',"&quot;")
    html = html.replace("'","&#39;")
    return html

#-------------------------------------------------------------------------
# Trimming: remove new-line, tabulation and spaces of the string
# Used to compute CRC32 on significant data only
def trim (str):
    str = str.replace(' ', '')
    str = str.replace('\s', '')
    str = str.replace('\t', '')
    str = str.replace('\r', '')
    str = str.replace('\n', '')
    return str

#-------------------------------------------------------------------------
# Return the highest bit-alignment of the given address
def whichDataAlignment(value):
    if (value % 2 != 0): return 8
    if (value % 4 != 0): return 16
    if (value % 8 != 0): return 32
    return 64


#-------------------------------------------------------------------------
# DATA ALIGNMENT depends on different conditions: PLC brand, model and also
# depends on type of data structure (scalar or array)
# These methods are used to compute the correct address of each register respecting
# this conditions.
# Attention!! $addr is the relative address of the register within the block including
# this alignment while $msize is still the useful memory of the data which does not include
# memory spaces because of alignment shifts of the following variables if any.
# This remark is not true for the block mem-size which naturally includes the alignment spaces.
#

# Adjust the register address relying on the SCHNEIDER Premium/Quantum alignment constraints
# Unity Premium/Quantum is 16bits processor and base-address is interpreted as 16bit address
# Byte elements of array (including STRING) use 8bit alignment.
def alignPremiumRegAddress(addr, format, size, dim, dim2, strLen):
    global msize

    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
    
    algnt  = 16
    while(whichDataAlignment(addr) < algnt):
        addr+=1
    msize = dim * dim2 * strLen * size    #compute memory data size
    if (format in ['int8', 'char']):   #8bit signed type use word alignment (16bit)
        msize = msize * 2            # while string and uint8-array use 8bit alignment
    return addr
    
# Adjust the register address relying on the SCHNEIDER M340 alignment constraints
# Unity M340 is 32bits processor but base-address is interpreted as 16bit address (required even address in the mapping)
# 32bits alignment except for 8/16bits (16bits alignment)
# Byte elements of array (including STRING) use 8bit alignment.
def alignM340RegAddress(addr, format, size, dim, dim2, strLen):
    global msize
    
    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
    
    algnt  = 16
    if (size > 2):
        algnt = 32
    while(whichDataAlignment(addr) < algnt):
        addr+=1
    msize = dim * dim2 * strLen * size    #compute memory data size
    if (format in ['int8', 'char']):   #8bit signed type use word alignment (16bit)
        msize = msize * 2            # while string and uint8-array use 8bit alignment
    return addr
    
# Adjust the register address relying on the DIGI-Rabbit RCMx alignment constraints
# Use  16bits processor and base-address is interpreted as 16bit address
def alignDIGIRegAddress(addr, format, size, dim, dim2, strLen):
    global msize

    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
    
    algnt  = 16
    while(whichDataAlignment(addr) < algnt):
        addr+=1
    msize = dim * dim2 * strLen * size    #compute memory data size
    if (format != 'string'):                       
        if (size == 1):                   #but it's a 8bit type
            msize = msize * 2        #so, it's a word alignment (only string use byte alignment)
    return addr

def alignCNVRegAddress(addr, format, size, dim, dim2, strLen):
    global regCounter
    return regCounter

# Adjust the register relying on the SIEMENS Simatic alignment constraints
# 8bits alignment except for array and >8bits data (16bits alignment)
# In case of string, its length has to be +2 to hold info on string
def alignSimaticRegAddress(addr, format, size, dim, dim2, strLen):
    global msize
    
    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
        
    algnt  = 8
    if ((size > 1) | (dim > 1) | (dim2 > 1) | (strLen > 1)):
        algnt = 16
    while(whichDataAlignment(addr) < algnt):
        addr+=1
    if strLen > 1:                          # register is a string, 
        if ((strLen % 2) != 0): strLen+=1   #adjusts the global size for byte type (word alignment)
        strLen+=2   #add increment to strLen first two bytes for info on string (len, maxlen)
          
    msize = dim * dim2 * strLen * size      #compute memory data size
    return addr

# Adjust the register address relying on the BECKHOFF Twincat BC9020 alignment constraints.
# TwinCAT BCxx PLCs are 16bits processor and base-address is interpreted as 16bit address
# 8bits alignment between elements of 8bit-data array (including strings).
# In case of string, its length has to be +1 to hold end-terminator.
def alignBCxxRegAddress(addr, format, size, dim, dim2, strLen):
    global msize

    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
    
    algnt  = 16
    while(whichDataAlignment(addr) < algnt):
        addr+=1
        
    if strLen > 1:          # register is a string, 
        strLen += 1       #TwinCAT  requires  '\0' string terminator.                             
    
    msize = dim * dim2 * strLen * size    #compute memory data size
    return addr

# Adjust the register address relying on the BECKHOFF Twincat CX9020 alignment constraints.
# TwinCAT CXxx PLCs are 32bits processor but base-address is interpreted as 16bit address (required even address in the mapping)
# 32bits alignment except for 8/16bits (16bits alignment).
# 8bits alignment between elements of 8bit-data array (including strings).
# In case of string, its length has to be +1 to hold end-terminator.
def alignCXxxRegAddress(addr, format, size, dim, dim2, strLen):
    global msize
    
    # internal string/unicode to integer cast
    size = int(size)
    addr = int(addr)
    dim = int(dim)
    dim2 = int(dim2)
    strLen = int(strLen)
    
    algnt  = 16
    if (size > 2):
        algnt = 32
    while(whichDataAlignment(addr) < algnt):
        addr+=1
        
    if strLen > 1:          # register is a string, 
        strLen += 1       #TwinCAT  requires  '\0' string terminator.
                                     
    msize = dim * dim2 * strLen * size    #compute memory data size
    return addr
    
#block-address = block DB-Number
def computeSiemensBlockBlkAddress(regAddr, classAddr, nbDev):
    #Compute block mem-size at first: each block has a dedicated DB-number which contains an array of devices
    #each element of the array is a structure which must be 16bits aligned!
    global blkMemSize
    global plcModel
    global blkAddr
    
    # internal string/unicode to integer cast
    regAddr = int(regAddr)
    classAddr = int(classAddr)
    nbDev = int(nbDev)
    
    nextBlockAddr = regAddr
    algnt = whichBaseAlignment[plcModel]    #get the base struct-alignment of that PLC (16bit here)
    # then adjust the next block-address by respecting this base-alignment.
    while(int(whichDataAlignment(nextBlockAddr)) < int(algnt)):
        nextBlockAddr+=1
    blkMemSize = nextBlockAddr
    # then compute the next block DB-number and return the current one
    tmpblkAddr = blkAddr    
    blkAddr = blkAddr + 1

    return tmpblkAddr + classAddr

#block-address = offset in the device memory
def computeSiemensDeviceBlkAddress(regAddr, classAddr, nbDev):
    #Compute block mem-size at first: each device has a dedicated DB-number which contains a sequence of blocks
    #each block is a structure which must be 16bits aligned!
    global blkMemSize
    global plcModel
    global blkAddr
    
    # internal string/unicode to integer cast
    regAddr = int(regAddr)
    classAddr = int(classAddr)
    nbDev = int(nbDev)

    nextBlockAddr = regAddr
    algnt  = whichBaseAlignment[plcModel]   #get the base struct-alignment of that PLC (16bit here)
    #then adjust the next block-address by respecting this base-alignment.                          
    while(int(whichDataAlignment(nextBlockAddr)) < int(algnt)):
        nextBlockAddr+=1
    blkMemSize = nextBlockAddr
    #then compute the next block DB-number and return the current one
    tmpblkAddr = blkAddr
    blkAddr = blkAddr + blkMemSize
    return tmpblkAddr
    
    
# Block-address = absolute address in the PLC memory
def computeSchneiderBlockBlkAddress(regAddr, classAddr, nbDev):
    global blkMemSize
    global plcModel
    global blkAddr

    # Internal string/unicode to integer cast
    regAddr = int(regAddr)
    classAddr = int(classAddr)
    nbDev = int(nbDev)

    # Compute block mem-size at first:
    # It corresponds to the next potential address that is 16bits or 32bits with SCHNEIDER.
    nextBlockAddr = regAddr
    
    # Each data-block for each class and device will be aligned on the worst-case alignment
    # of the given PLC. This logic allows having a unique address computing whatever the mode (DEVICE/BLOCK)
    # and the PLC model (16bits or 32bits).
    algnt  = whichBaseAlignment[plcModel] # get the "worst-case" base-alignment of that PLC
    #then adjust the next block-address by respecting this base-alignment.                          
    while(whichDataAlignment(nextBlockAddr) < int(algnt)):
        nextBlockAddr+=1
    blkMemSize = nextBlockAddr  
    # then compute the next block DB-number and return the current one
    tmpblkAddr = blkAddr   
    # next absolute address: relies on the global class-address
    blkAddr = blkAddr + (nbDev * blkMemSize)
    return classAddr + tmpblkAddr
    
    
# Block-address = offset in the device memory
def computeSchneiderDeviceBlkAddress(regAddr, classAddr, nbDev):
    global blkMemSize
    global plcModel
    global blkAddr
        
    # internal string/unicode to integer cast
    regAddr = int(regAddr)
    classAddr = int(classAddr)
    nbDev = int(nbDev)
    
    #Compute block mem-size at first:
    #it corresponds to the next potential address that is 16bits or 32bits with SCHNEIDER.
    nextBlockAddr = regAddr
    #Each data-block for each class and device will be aligned on the worst-case alignment
    #of the given PLC. This logic allows having a unique adress computing whatever the mode (DEVICE/BLOCK)
    #and the PLC model (16bits or 32bits).
    algnt  = whichBaseAlignment[plcModel] #get the "worst-case" base-alignment of that PLC
    #then adjust the next block-address by respecting this base-alignment.          
    while(whichDataAlignment(nextBlockAddr) < int(algnt)):
        nextBlockAddr+=1
    blkMemSize = nextBlockAddr
    #then compute the next block DB-number and return the current one
    tmpblkAddr = blkAddr   
    #next relative address: independent from the global class-address
    blkAddr = blkAddr + blkMemSize
    return tmpblkAddr
    
def computeNiDeviceBlkAddress(regAddr, classAddr, nbDev):
    global blockCounter
    blockCounter = blockCounter + 1
    return blockCounter

#-------------------------------------------------------------------------
# Compute the next register address relying on the data start-address and size.
def computeAnyNextRegAddress(brand, addr, dim, dim2=1):
    global msize
    
    # internal string/unicode to integer cast
    addr = int(addr)
    dim = int (dim)
    dim2 = int(dim2)
    
    addr = addr + msize;    #compute the next address for any case
    if (brand == 'SIEMENS'):
        # SIEMENS requires specific treatment in case of array.
        if ((dim > 1) | (dim2 > 1)):
            #SIEMENS array is always followed by 16bits adressing
            algnt = 16
            while(whichDataAlignment(addr) < algnt):
                addr+=1
    elif (brand == 'NI'):
        msize = 0
    return addr


#-------------------------------------------------------------------------
def computeAnyBlockInstAddress(classAddr, devSize):
    global instAddr
    tmpInstAddr = instAddr
    instAddr = instAddr + 1     #device-index
    return tmpInstAddr

def computeSiemensDeviceInstAddress(classAddr, devSize):
    global instAddr 
    tmpInstAddr = instAddr
    instAddr = instAddr + 1     #device DB-number
    return tmpInstAddr + classAddr

def computeSchneiderDeviceInstAddress(classAddr, devSize):
    global instAddr
    tmpInstAddr = instAddr
    instAddr = instAddr + devSize #absolute address
    return tmpInstAddr + classAddr

def computeNiDeviceInstAddress(classAddr, devSize):
    return 0
#-------------------------------------------------------------------------
# Compute the base DB-number of the next class in the SIEMENS PLC memory
# BLOCK mode requires 1 DB per class block
def computeSiemensBlockNextBaseAddress(classAddr, nbBlk, nbDev, devSize):
    global classMem
    global plcLast
    global plcSize

    # internal string/unicode to integer cast
    classAddr   = int(classAddr)
    nbBlk       = int(nbBlk)
    nbDev       = int(nbDev)
    devSize     = int(devSize)
    
    byteSize = nbDev * devSize
    plcSize = int(plcSize) + int(byteSize)  #global size of the plc configuration
    startDB = classAddr                     #first DB used for this class
    plcLast = classAddr + nbBlk - 1         #last DB used
    classMem = "DB%d..DB%d / %d bytes"%(startDB,plcLast,byteSize)
    return plcLast + 1                      #next DB number

# Compute the base DB-number of the next instance in the SIEMENS PLC memory
# DEVICE mode requires 1 DB per class instance
def computeSiemensDeviceNextBaseAddress(classAddr, nbBlk, nbDev, devSize):
    global classMem
    global plcLast
    global plcSize
    classAddr   = int(classAddr)
    nbBlk       = int(nbBlk)
    nbDev       = int(nbDev)
    devSize     = int(devSize)
    
    byteSize = nbDev * devSize
    plcSize = plcSize + byteSize            #global size of the plc configuration
    startDB = classAddr                     #first DB used for this class
    plcLast = classAddr + nbDev - 1         #last DB used
    classMem = "DB%d..DB%d / %d bytes"%(startDB,plcLast,byteSize)
    return plcLast + 1                      #next DB number

# Compute the base-address of the next class in the SCHNEIDER PLC memory
# DEVICE or BLOCK mode use the same memory size: next base address should be the same
# 'used-mem' info is expressed in words (/2) but address is computed in bytes
def computeSchneiderAnyNextBaseAddress(classAddr, nbBlk, nbDev, devSize):
    global classMem
    global plcLast
    global plcSize
    classAddr   = int(classAddr)
    nbBlk       = int(nbBlk)
    nbDev       = int(nbDev)
    devSize     = int(devSize)
    
    wordSize = (nbDev * devSize) / 2
    plcSize = plcSize + wordSize            #global size of the plc configuration
    startAddr = classAddr / 2               #first word address used for this class
    plcLast = startAddr + wordSize - 1      #LAST word address used for this class
    classMem = "MW%d..MW%d / %d words"%(startAddr,plcLast,wordSize);
    return (classAddr + (wordSize * 2))     #next word address expressed in bytes

def computeNiAnyNextBaseAddress(classAddr, nbBlk, nbDev, devSize):
    return 0

#-------------------------------------------------------------------------
# computeChecksumController
# checksum procedure:
#     CRC = Major-number + PLC-base-address
#     For each PLC-Class (cluster) of the controller mapping:
#         CRC += class-name + class-version + class-access //DEVICE/BLOCK
#         For each Block
#             CRC += block-name + block-mode //R/W
#             For each Register
#                 CRC += reg-name + reg-format + array-dim1 + array-dim2 + (string-length) optional
#         CRC += nb-instance
#
def computeChecksumController( workspacePath, controllerNode, silecsVersion, PLCbaseAddress, funcGetSilecsDesignFilePath,  logTopics={'errorlog': True}):
    majorSilecsVersion = iecommon.getMajorSilecsVersion(silecsVersion)
    CRC32 = zlib.crc32(trim(str(majorSilecsVersion)),0)& 0xffffffff
    CRC32 = zlib.crc32(trim(str(PLCbaseAddress)),CRC32)& 0xffffffff
    for silecsDesign in controllerNode.xpathEval("SilecsDesign"):
        CRC32       = zlib.crc32(trim(silecsDesign.prop("silecs-design-name")),CRC32)& 0xffffffff
        CRC32       = zlib.crc32(trim(silecsDesign.prop("silecs-design-version")),CRC32)& 0xffffffff
        designDOM = loadSilecsDesignDOM(workspacePath, silecsDesign, silecsVersion, funcGetSilecsDesignFilePath)
        CRC32 = computeChecksumClass(designDOM,CRC32,logTopics)
        for device in silecsDesign.xpathEval("Device"):
            CRC32     = zlib.crc32(trim(device.prop("device-name")),CRC32)& 0xffffffff
    iecommon.logInfo("CRC32: %s" % str(CRC32),logTopics)
    return CRC32
    
def computeChecksumClass(designDOM, CRC32, logTopics={'errorlog': True}):
    blockList = designDOM.xpathEval("Block")
    for blockNode in blockList:
        blockName = blockNode.prop('name')
        blockMode = blockNode.prop('mode')
        registerList = blockNode.xpathEval("Register")
        for register in registerList:
            regName   = register.prop("name")
            CRC32     = zlib.crc32(trim(regName),CRC32)& 0xffffffff

            regFormat = register.prop("format")
            CRC32     = zlib.crc32(trim(regFormat),CRC32)& 0xffffffff

            try:
                # array-dim1 is an optional attribute!!!
                regDim1   = register.prop("array-dim1")
            except:
                regDim1   = "1"
            CRC32     = zlib.crc32(trim(regDim1),CRC32)& 0xffffffff

            try:
                # array-dim2 is an optional attribute!!!
                regDim2   = register.prop("array-dim2")
            except:
                regDim2   = "1"
            CRC32     = zlib.crc32(trim(regDim2),CRC32)& 0xffffffff

            try:
                # string-len is an optional attribute!!!
                stringLength = register.prop("string-len")
                CRC32     = zlib.crc32(trim(stringLength),CRC32)& 0xffffffff
            except:
                pass
    return CRC32
    
#-------------------------------------------------------------------------
# Hash table
#-------------------------------------------------------------------------

whichPLCBrand = {
    'SIMATIC_S7-300'    : 'SIEMENS',
    'SIMATIC_S7-400'    : 'SIEMENS',
    'SIMATIC_S7-1200'   : 'SIEMENS',
    'SIMATIC_S7-1500'   : 'SIEMENS',
    'SIMATIC_ET-200S'   : 'SIEMENS',
    'SIMATIC_S7-VIRTUAL': 'SIEMENS',
    'Premium'           : 'SCHNEIDER',
    'Quantum'           : 'SCHNEIDER',
    'M340'              : 'SCHNEIDER',
    'Compact_RIO'       : 'NI',
    'PXI_RT'            : 'NI',
    'PXI_Windows'       : 'NI',
    'PC_Windows'        : 'NI',
    'Other_Support_CNV' : 'NI',
    'Rabbit_RCM_4010'   : 'DIGI',
    'Rabbit_RCM_2000'   : 'DIGI',
    'BC9020'            : 'BECKHOFF',
    'CX9020'            : 'BECKHOFF'
}

# The following constant are used to align the address of a device block {set of registers}.
# With SCHNEIDER PLC, block is aligned depending on his 'worst-case' alignement {16bits or 32bits}.
# This logic allows having a unique adress computing whatever the mode {DEVICE/BLOCK} and the PLC model.
# Concerning SIEMENS PLC, block alignment should respect the 16bits alignement constraint for Struct&Array.
whichBaseAlignment = {
    'SIMATIC_S7-300'    : '16',
    'SIMATIC_S7-400'    : '16',
    'SIMATIC_S7-1200'   : '16',
    'SIMATIC_S7-1500'   : '16',
    'SIMATIC_ET-200S'   : '16',
    'SIMATIC_S7-VIRTUAL': '16',
    'Premium'           : '16',
    'Quantum'           : '16',
    'M340'              : '32',     # any data-block of M340 model will start on 32bits address
                                  # even if the first register of the block is a 16bit data.
    'Rabbit_RCM_4010' : '16',
    'Rabbit_RCM_2000' : '16',
    'BC9020'          : '16',
    'CX9020'          : '32'                          
}

whichDataSize = {
    'uint8'        : '1',
    'int8'          : '1',
    'uint16'      : '2',
    'int16'        : '2',
    'uint32'      : '4',
    'int32'        : '4',
    'float32'     : '4',
    'uint64'      : '8',
    'int64'        : '8',
    'float64'     : '8',
    'string'       : '1',
    'date'         : '8',
    # deprecated formats
    'char'         : '1',
    'byte'         : '1',
    'word'        : '2',
    'dword'      : '4',
    'int'            : '2',
    'dint'          : '4',
    'real'          : '4',
    'dt'             : '8'
}

whichRegAddressFunction = {
    'SIMATIC_S7-300'          : alignSimaticRegAddress,
    'SIMATIC_S7-400'          : alignSimaticRegAddress,
    'SIMATIC_S7-1200'         : alignSimaticRegAddress,
    'SIMATIC_S7-1500'         : alignSimaticRegAddress,
    'SIMATIC_ET-200S'         : alignSimaticRegAddress,
    'SIMATIC_S7-VIRTUAL'      : alignSimaticRegAddress,
    'Premium'                 : alignPremiumRegAddress,
    'Quantum'                 : alignPremiumRegAddress,
    'M340'                    : alignM340RegAddress,
    'Compact_RIO'             : alignCNVRegAddress,
    'PXI_RT'                  : alignCNVRegAddress,
    'PXI_Windows'             : alignCNVRegAddress,
    'PC_Windows'              : alignCNVRegAddress,
    'Other_Support_CNV'       : alignCNVRegAddress,
    'Rabbit_RCM_4010'         : alignDIGIRegAddress,
    'Rabbit_RCM_2000'         : alignDIGIRegAddress,
    'BC9020'                  : alignBCxxRegAddress,
    'CX9020'                  : alignCXxxRegAddress
}

whichBlkAddressFunction = {
    'SIEMENS'+'BLOCK_MODE'        : computeSiemensBlockBlkAddress,
    'SIEMENS'+'DEVICE_MODE'       : computeSiemensDeviceBlkAddress,
    'SCHNEIDER'+'BLOCK_MODE'    : computeSchneiderBlockBlkAddress,
    'SCHNEIDER'+'DEVICE_MODE'   : computeSchneiderDeviceBlkAddress,
    'NI'+'DEVICE_MODE'                  : computeNiDeviceBlkAddress,
    'DIGI'+'BLOCK_MODE'           : computeSchneiderBlockBlkAddress,
    'DIGI'+'DEVICE_MODE'          : computeSchneiderDeviceBlkAddress,
    'BECKHOFF'+'BLOCK_MODE'      : computeSchneiderBlockBlkAddress
}

whichInstAddressFunction = {
    'SIEMENS'+'BLOCK_MODE'         : computeAnyBlockInstAddress,
    'SIEMENS'+'DEVICE_MODE'        : computeSiemensDeviceInstAddress,
    'SCHNEIDER'+'BLOCK_MODE'    : computeAnyBlockInstAddress,
    'SCHNEIDER'+'DEVICE_MODE'   : computeSchneiderDeviceInstAddress,
    'NI'+'DEVICE_MODE'                  : computeNiDeviceInstAddress,
    'DIGI'+'BLOCK_MODE'           : computeAnyBlockInstAddress,
    'DIGI'+'DEVICE_MODE'          : computeSchneiderDeviceInstAddress,
    'BECKHOFF'+'BLOCK_MODE'      : computeAnyBlockInstAddress
}

whichBaseAddressFunction = {
    'SIEMENS'+'BLOCK_MODE'         : computeSiemensBlockNextBaseAddress,
    'SIEMENS'+'DEVICE_MODE'        : computeSiemensDeviceNextBaseAddress,
    'SCHNEIDER'+'BLOCK_MODE'     : computeSchneiderAnyNextBaseAddress,
    'SCHNEIDER'+'DEVICE_MODE'    : computeSchneiderAnyNextBaseAddress,
    'NI'+'DEVICE_MODE'                   : computeNiAnyNextBaseAddress,
    'DIGI'+'BLOCK_MODE'            : computeSchneiderAnyNextBaseAddress,
    'DIGI'+'DEVICE_MODE'           : computeSchneiderAnyNextBaseAddress,
    'BECKHOFF'+'BLOCK_MODE'       : computeSchneiderAnyNextBaseAddress
}

#Base address provided by the user in the Deployment relies on the PLC model:
#Unity Premium/Quantum is interpreted as 16bit address
#Unity M340 is interpreted as 32bit address
#Factor is used to convert the base-address to byte-addressing for computation.
whichAddressFactor = {      
    'SIMATIC_S7-300'    : 1,  # Simatic does not use absolute addressing {DB-number}
    'SIMATIC_S7-400'    : 1,
    'SIMATIC_S7-1200'   : 1,
    'SIMATIC_S7-1500'   : 1,
    'SIMATIC_ET-200S'   : 1,
    'SIMATIC_S7-VIRTUAL': 1,
    'Premium'           : 2,  # Premium/Quantum uses 16bit adressing for 32bit values
    'Quantum'           : 2,
    'M340'              : 2,
    'Compact_RIO'       : 0,
    'PXI_RT'            : 0,
    'PXI_Windows'       : 0,
    'PC_Windows'        : 0,
    'Other_Support_CNV' : 0,
    'Rabbit_RCM_4010'   : 2,
    'Rabbit_RCM_2000'   : 2,
    'BC9020'            : 2,    # BC90xx use 16 bit addressing
    'CX9020'            : 1
} 

def loadSilecsDesignDOM(workspacePath, classNode, silecsVersion, funcGetSilecsDesignFilePath):
    if classNode.prop("silecs-design-name") == "SilecsHeader":
        silecsHeader = xmltemplate.getSilecsHeader(silecsVersion)
        return libxml2.parseDoc(silecsHeader)
    else:
        designPath = funcGetSilecsDesignFilePath(workspacePath, classNode.prop("silecs-design-name"))
        return libxml2.parseFile(designPath)

# Needed to encapsulate "genParam" in order to allow unit-testing (fake all file interactions)
def genParamBase( funcGetSilecsDesignFilePath, funcGetParameterFile, funcGetSilecsDeployFilePath,  funcGetParameterFileDirectory, workspacePath, deployName, deployVersion, silecsVersion, logTopics={'errorlog': True}):
    # Global variable links
    global plcModel, plcBrand, plcSystem, plcProtocol, plcSize, plcLast
    global PLCbaseAddress, checksumRef, owner, deviceMemSize, blkAddr, nbBlock, msize
    global blkMemSize
    global instAddr, classAddr
    global classMem
    
    global blockCounter     # used for NI block address generation
    global regCounter   # used for NI register address generation

    # Check the Deployment document exist for that PLC
    deployPath = funcGetSilecsDeployFilePath(workspacePath, deployName)
    if(not os.path.isfile(deployPath)):
        iecommon.logError(deployName + "deployment file cannot be found in provided workspace",logTopics)
    
    # Create the Deployment DOM object
    deployDOM = libxml2.parseFile(deployPath)
    
    # Check that Deployment data are consistent
    deployUnitNode = deployDOM.xpathEval("/SILECS-Deploy/Deploy-Unit")[0] # only one Deploy-Unit
    
    if (deployUnitNode.prop('name') != deployName):
        iecommon.logError("Deployment name is not consistent", True,logTopics)

    #-------------------------------------------------------------------------
    # Extract information from the XML
    #-------------------------------------------------------------------------
    
    # Extract Deployment owner from the Information node
    owner = deployDOM.xpathEval("/SILECS-Deploy/Information/Owner")[0].prop('user-login')

    #-------------------------------------------------------------------------
    # SECOND pass on each Design documents: Generate parameters
    # For each class generate XML-ouput document relying on Design & Deployment
    #-------------------------------------------------------------------------
    iecommon.logDebug("------Processing for output file generation------",logTopics)
    paramPath = funcGetParameterFileDirectory(workspacePath, deployName)

    #create output directory if necessary
    if not os.path.exists(paramPath):
        os.makedirs(paramPath)
        iecommon.logDebug("create directory %s" %paramPath,logTopics)
    
    iecommon.addSilecsHeaderToClasses(deployDOM)
    controllerNodes = deployDOM.xpathEval("/SILECS-Deploy/Controller")
    for controllerNode in controllerNodes:
        plcNode  = controllerNode.xpathEval("*[@system]")[0]
        # Extract PLC global information
        plcSystem     = plcNode.prop('system')
        plcModel        = plcNode.prop('model')
        plcBrand        = whichPLCBrand[plcModel]
        plcProtocol     = plcNode.prop('protocol')
        addressFactor   = whichAddressFactor[plcModel]

        if plcBrand == 'SIEMENS':
            PLCbaseAddress = long(plcNode.prop('base-DB-number'))
            
        elif plcBrand == 'SCHNEIDER':
            PLCbaseAddress = long(plcNode.prop('base-address'))
            PLCbaseAddress = PLCbaseAddress * addressFactor
            
        elif plcBrand == 'BECKHOFF':
            PLCbaseAddress = long(plcNode.prop('base-address'))
            # Use plcModel to set appropriate initial offset
            if plcModel == 'BC9020':
                offset = 32768
            elif plcModel == 'CX9020':
                offset = 24576
            PLCbaseAddress = (PLCbaseAddress * addressFactor) + offset
                
        elif plcBrand == 'DIGI':
            PLCbaseAddress = long(plcNode.prop('base-address'))
            PLCbaseAddress = PLCbaseAddress * addressFactor
            
        else: # plcBrand == 'NI'
            PLCbaseAddress = 0

        classBaseAddress = PLCbaseAddress

        # Messagges for debugging purpose
        iecommon.logDebug("------ XML extracted informations ------",logTopics)
        iecommon.logDebug("owner = "+owner,logTopics)
        iecommon.logDebug("plcSystem = "+plcSystem,logTopics)
        iecommon.logDebug("plcModel = "+plcModel,logTopics)
        iecommon.logDebug("plcBrand = "+plcBrand,logTopics)
        iecommon.logDebug("plcProtocol = "+plcProtocol,logTopics)
        iecommon.logDebug("addressFactor = %d" %addressFactor,logTopics)
        iecommon.logDebug("PLCbaseAddress = %d" %PLCbaseAddress,logTopics)

        paramFile = funcGetParameterFile(workspacePath, deployName, controllerNode.prop('host-name'))
        iecommon.logInfo("Generate xml for parameters file: "+paramFile,logTopics)

        # Create the parameter DOM for the output file
        paramDOM  = libxml2.newDoc(version='1.0')
        outputRoot  = libxml2.newNode('SILECS-Param')
        outputRoot.setProp("silecs-version", silecsVersion)
        paramDOM.addChild(outputRoot)

        #-------------------------------------------------------------------------
        # Generate section <Mapping-Info></Mapping-Infon>
        #-------------------------------------------------------------------------
        paramMappingInfoNode = libxml2.newNode("Mapping-Info")
        outputRoot.addChild(paramMappingInfoNode)

        # Add Owner
        paramOwnerNode = libxml2.newNode('Owner')
        paramOwnerNode.setProp("user-login", owner)
        paramMappingInfoNode.addChild(paramOwnerNode)
        
        # Add Generation
        element2 = libxml2.newNode('Generation')
        currentDate = str(datetime.datetime.now())
        element2.setProp("date",currentDate)
        paramMappingInfoNode.addChild(element2)
        
        # Add Deployment
        element2 = libxml2.newNode('Deployment')
        CRC32 = computeChecksumController(workspacePath, controllerNode, silecsVersion, PLCbaseAddress, funcGetSilecsDesignFilePath, logTopics)
        element2.setProp("checksum", str(CRC32))
        paramMappingInfoNode.addChild(element2)

        #-------------------------------------------------------------------------
        # Generate section <SILECS-Mapping></SILECS-Mapping>
        #-------------------------------------------------------------------------
        paramSilecsMappingNode = libxml2.newNode("SILECS-Mapping")
        paramSilecsMappingNode.setProp("plc-name", controllerNode.prop('host-name'))
        paramSilecsMappingNode.setProp("plc-brand", plcBrand)
        paramSilecsMappingNode.setProp("plc-system", plcSystem)
        paramSilecsMappingNode.setProp("plc-model", plcModel)
        paramSilecsMappingNode.setProp("protocol", plcProtocol)
        paramSilecsMappingNode.setProp("address", str(PLCbaseAddress))
        paramSilecsMappingNode.setProp("domain", "NotUsed")
        paramSilecsMappingNode.setProp("used-mem", "TODO")
        outputRoot.addChild(paramSilecsMappingNode)

        classNodes = controllerNode.xpathEval('SilecsDesign')
        for classNode in classNodes:
            iecommon.logDebug("-----------------------------------------",logTopics)
            iecommon.logDebug("------  Analysing Class %s   ------"%classNode.prop("silecs-design-name"),logTopics)
            iecommon.logDebug("-----------------------------------------",logTopics)
            
            designDOM = loadSilecsDesignDOM(workspacePath, classNode, silecsVersion, funcGetSilecsDesignFilePath)

            # Extract the number of devices ------------------------------------------

            deviceLabelList=classNode.xpathEval('Device')
            nbDevice = len(deviceLabelList)
            iecommon.logDebug("Class %s uses device-list and has %s devices" %(classNode,nbDevice),logTopics)
            
            paramClassNode = libxml2.newNode("SILECS-Class")
            paramSilecsMappingNode.addChild(paramClassNode)
            paramClassNode.setProp("name", classNode.prop("silecs-design-name"))
            paramClassNode.setProp("version", classNode.prop("silecs-design-version"))
            
            #-------------------------------------------------------------------------
            # Generate section <Block></Block>
            #-------------------------------------------------------------------------
            blockCounter = 0
            
            deviceMemSize = 0   # global memory-size of one class instance (sum of block-memory size)
            blkAddr = 0                 # memory address of the block (using byte addressing)
            nbBlock = 0             # number of block of the class

            computeBlkAddress = whichBlkAddressFunction[plcBrand + plcProtocol]
            
            # INNER LOOP TO ACCESS AT BLOCK LEVEL (LOOP-2)
            for block in designDOM.xpathEval("//Block"):    #LOOP-2
                element3 = libxml2.newNode("Block")
                element3.setProp("name", block.prop("name"))
                element3.setProp("mode", block.prop("mode"))
                iecommon.logDebug("----- Processing Block "+block.prop('name')+" from design -----",logTopics)
                # Initialization for inner loop local variables 
                regAddress = 0      # memory address of the register (using byte addressing)
                blockSize = 0           # block size (sum of the register size)
                blockMemSize = 0    # block size (sum of the regist)
                instAddr = 0                # initial class address BUG FIXED
                regCounter = 0      # used for NI register address generation
                alignRegAddress     = whichRegAddressFunction [plcModel]
        
                #-------------------------------------------------------------------------
                # Generate section <Register></Register>
                #-------------------------------------------------------------------------
                # INNER LOOP TO ACCESS AT REGISTER LEVEL (LOOP-3)
                for registerNode in block.xpathEval("Register"):
                    designRegister = Register(registerNode)
                    iecommon.logDebug("------ Processing Register "+designRegister.name+ " ------")
                    # Set length attribute only for string registers
                    if designRegister.format == 'string':
                        if plcBrand == 'RABBIT':
                            # RABBIT has 8bit memory alignment but uses 16bit word communication protocol (MODBUS).
                            # String buffer (8bit elements) must have even number of bytes.
                            if ((int(designRegister.dim1)*int(designRegister.dim2)*int(designRegister.stringLength)) % 2 == 1): #XML CliX cannot check that constraint!   
                                iecommon.logError("String length of %s designRegister must be an even value." %designRegister.name, False,logTopics)
                                sys.exit(2)
                    
                    # Set register size
                    regSize = whichDataSize[designRegister.format]
                    # Set register address
                    regAddress = alignRegAddress(regAddress, designRegister.format, regSize, designRegister.dim1, designRegister.dim2, designRegister.stringLength)
                    regCounter = regCounter + 1 # used for NI register address generation
                    # Set register mem-size
                    paramRegister = ParamRegister()
                    paramRegister.initWithDesignRegister(designRegister,regSize,regAddress,msize)
                    # Compute address for the next register
                    regAddress = computeAnyNextRegAddress(plcBrand, regAddress, designRegister.dim1, designRegister.dim2)
                    # Append register
                    element3.addChild(paramRegister.xmlNode)
                    #iterativelly compute the block size (accumulator initialized outside the loop)
                    blockSize = blockSize + (int(regSize) * int(designRegister.dim1) * int(designRegister.dim2))
                # END OF INNER LOOP TO ACCESS AT REGISTER LEVEL (LOOP-3)

                # Set block size
                element3.setProp("size", str(blockSize))
                # Set block address
                element3.setProp("address",str(computeBlkAddress(regAddress, int(classBaseAddress) , nbDevice)))
                # Set block mem-size
                element3.setProp("mem-size", str(blkMemSize))
                # Append block
                paramClassNode.addChild(element3)
                # Count the number of devices
                nbBlock = nbBlock+1
                # Accumulate blkMemSize to compute the total deviceMemSize
                deviceMemSize = deviceMemSize + blkMemSize

            # END OF INNER LOOP TO ACCESS AT BLOCK LEVEL (LOOP-2)
            # Set block Address
            paramClassNode.setProp("address", str(classBaseAddress))
            
            #-------------------------------------------------------------------------
            # Devices
            #-------------------------------------------------------------------------
            computeInstAddress = whichInstAddressFunction[ plcBrand + plcProtocol ]

            for device in deviceLabelList:
                element3 = libxml2.newNode("Instance")     
                element3.setProp("label", device.prop("device-name"))
                element3.setProp("address", str(computeInstAddress(classBaseAddress, deviceMemSize)))
                paramClassNode.addChild(element3)

            # Compute the memory address for the next class
            computeBaseAddress = whichBaseAddressFunction [plcBrand+plcProtocol]
            classBaseAddress = computeBaseAddress(classBaseAddress, nbBlock, nbDevice, deviceMemSize);
            
            # Set class used-memory
            paramClassNode.setProp("used-mem", str(classMem))
            iecommon.logInfo("Used-memory for Class "+classNode.prop("silecs-design-version")+": "+str(classMem),logTopics)

        #------------------------------------------------------
        # Generate output XML
        #------------------------------------------------------
        iefiles.saveXMLToFile(paramFile,paramDOM)
        
        iecommon.logDebug("---------------------------------------------------------",logTopics)
        iecommon.logDebug("------GENERATED FILE: %s "%paramFile,logTopics)
        iecommon.logDebug("---------------------------------------------------------",logTopics)
     
    # GSI-Fix there are more then two types now ... better say nothing than something which could be wrong for the else-case
    #Display details about the PLC memory using for the global configuration
    #if (plcBrand == 'SIEMENS'):
    #    # SIEMENS works with data-block addressing
    #    plcUsedMem = "DB"+str(PLCbaseAddress)+"..DB"+str(plcLast)+" / "+str(plcSize)+" bytes"
    #else:
    #    # SCHNEIDER works with absolute addressing
    #    startAddr   = PLCbaseAddress /2         #Memory info uses word-addressing
    #    lastAddr    = plcLast               #Already computed as byte-address
    #    plcUsedMem  = "MW"+str(startAddr)+"..MW"+str(lastAddr)+" / "+str(plcSize)+" words"
    
    # print plc used memory
    # iecommon.logInfo("Used-memory for PLC "+deployName+": "+str(plcUsedMem),logTopics)


#=========================================================================
# Entry point
#=========================================================================
def genParam(workspacePath, deployName, deployVersion, silecsVersion, logTopics={'errorlog': True}):
    genParamBase( iefiles.getSilecsDesignFilePath, iefiles.getParameterFile, iefiles.getSilecsDeployFilePath,  iefiles.getParameterFileDirectory, workspacePath, deployName, deployVersion, silecsVersion, logTopics)


