#!/usr/bin/python
# Copyright 2016 CERN and GSI
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

#=========================================================================
# Global definition
#=========================================================================

#-------------------------------------------------------------------------
# Hash-Table

whichRabbitFormat = {
	'uint8'   : 'uint16_t',
	'int8'    : 'int16_t',
	'uint16'  : 'uint16_t',
	'int16'   : 'int16_t',
	'uint32'  : 'uint32_t',
	'int32'   : 'int32_t',
	'float32' : 'float',
	'uint64'  : '8',
	'int64'   : '8',
    'string'  : 'uint8_t',    
#	'float64' : '8',    not supported by microcontroller
	'date'    : 'dt',
   	'char'	:'int16_t',
    'byte'	:'uint16_t',
    'word'	:'uint16_t',
    'dword'	:'uint32_t',
    'int'	:'int16_t',
    'dint'	:'int32_t',
    'real'	:'float',
    'dt'	:'dt'
}

#=========================================================================
# Utils
#=========================================================================

def toWordArray(strg):
    wordArray = ''
    for i in range(0, len(strg), 2):
        lsb = ord(strg[i])
        try: hsb = ord(strg[i+1]) 
        except: hsb = 0
        wordArray += "0x%02x%02x," %(hsb,lsb)
    return wordArray

#=========================================================================
# C Source template (.c file)
#=========================================================================

header = """
/* +-------------------------------------------------------------------
 * | Copyright CERN 2015
 * | SILECS - BE/CO-SRC
 * | April 2015
 * +-------------------------------------------------------------------
 *
 * Release : SILECS_%s
 *
 * The following code has been automatically generated by SILECS.
 *
 * N.B: This file relies on the existence of explicit C data type such
 * as int8_t, uint8_t, int16_t, etc....
 * If your compiler does not support them natively please implement 
 * them by including the data type definition provided on the SILECS
 * web page before including this header file.
 */

#define MODBUS_START_ADDRESS %s

/*---------------------------------------------------------------------
 * DT 
 * data type definition and related utilities
 *---------------------------------------------------------------------
 */

typedef struct
{
	uint8_t sc_100;     // second cent
	uint8_t sc;			// second
	uint8_t mn;         // minute
	uint8_t hh;         // hour
	uint8_t dd;         // day
	uint8_t mm;         // month
	uint8_t yy1;        // year
	uint8_t yy2;        // year2
} dt;

#define _frombcd(a) (int)(((a>>4)*10)+(a&0x0F))
#define _tobcd(a)   (((unsigned char)((a)/10)<<4)+((a)%%10))

void SILECS_set_dt(int8_t sc_100 ,int8_t sc, int8_t mn,int8_t hh,int8_t dd,int8_t mm,int32_t yy, dt *date)
{
	date->sc_100 = sc_100;
	date->sc     = _tobcd(sc);
	date->mn     = _tobcd(mn);
	date->hh     = _tobcd(hh);
	date->dd     = _tobcd(dd);
	date->mm     = _tobcd(mm);
	date->yy2    = _tobcd((int8_t)(yy/100));
	date->yy1    = _tobcd((int8_t)(yy%%100));
}

"""

#=========================================================================
# DATA TYPE DEFINITION
#=========================================================================

firstBlockUDT = """
/*---------------------------------------------------------------------
 * %s / v%s
 * BLOCK Type definition
 *---------------------------------------------------------------------
 */
"""

blockUDT = """
typedef struct 
{
%s
} _%s_%s;
"""

regScalar = """    %s %s;
"""

regArray  = """    %s %s[%s];
"""

regArray2d  = """    %s %s[%s][%s];
"""

stringArray = """    %s %s[%s][%s][%s];
"""

#=========================================================================
# INSTANTIATION
#=========================================================================

allocationComment = """
/*---------------------------------------------------------------------
 * MEMORY ALLOCATION
 * PROTOCOL: %s
 *---------------------------------------------------------------------
 */
"""

NBdeviceDefinition = """#define NB_%s_DEVICE %s
"""

#======= DEVICE MODE =================

deviceModeBlockInstantiation = """        _%s_%s %s;
"""

#deviceModeClass_deviceNumber = """
#    struct {
#        _%s_%s %s;
#    } %s_device[%s];
#"""

deviceModeClass_deviceNumber = """
    struct {
%s
    } %s_device[NB_%s_DEVICE];
"""

deviceModeClass_deviceList = """
    struct {
%s
    } %s;
"""

deviceMode_dataInstantiation = """
typedef struct {
    %s
} _SILECS_DATA_SEGMENT;

#define SILECS_DATA_SEGMENT_MODBUS_SIZE (sizeof(_SILECS_DATA_SEGMENT)/2)

union silecsData {
    _SILECS_DATA_SEGMENT data;
    uint16_t array[SILECS_DATA_SEGMENT_MODBUS_SIZE];
} silecsData;

"""

#======= BLOCK MODE =================

blockModeDeviceInstantiation_deviceNumber = """        _%s_%s device[NB_%s_DEVICE];
"""

blockModeDeviceInstantiation_deviceList = """        _%s_%s %s;
"""

blockModeBlockInstantiation = """
    struct {
%s    } %s_%s;
"""

blockMode_dataInstantiation = """
typedef struct {
    %s
} _SILECS_DATA_SEGMENT;

#define SILECS_DATA_SEGMENT_MODBUS_SIZE (sizeof(_SILECS_DATA_SEGMENT)/2)

union modbus_data {
    _SILECS_DATA_SEGMENT data;
    uint16_t array[SILECS_DATA_SEGMENT_MODBUS_SIZE];
} silecsData;

"""

#=========================================================================
# init
#=========================================================================

globalAllocation = """
    SILECS_INSTANCE_DATA data;
    uint16_t modbus_data = & data;
"""

initFunctionDeviceMode = """
/* Initialization function */
int SILECS_init()
{
	/* Silecs version initialization */
	strcpy((unsigned char *)silecsData.data.SilecsHeader_device[0].hdrBlk._version, "%s");
	
	/* Silecs checksum initialization */
	silecsData.data.SilecsHeader_device[0].hdrBlk._checksum  = %s;
	
	/* Silecs user initialization */
	strcpy((unsigned char *)silecsData.data.SilecsHeader_device[0].hdrBlk._user, "%s");
	
	/* Silecs date initialization */
	SILECS_set_dt(%s,%s,%s,%s,%s,%s,%s,&silecsData.data.SilecsHeader_device[0].hdrBlk._date);
}

"""

#initFunctionBlockMode = """
initFunctionBlockMode = """
/* Initialization function */
int SILECS_init()
{
	/* Silecs version initialization */
	strcpy((unsigned char *)silecsData.data.SilecsHeader_hdrBlk.device[0]._version, "%s");
	
	/* Silecs checksum initialization */
	silecsData.data.SilecsHeader_hdrBlk.device[0]._checksum  = %s;
	
	/* Silecs user initialization */
	strcpy((unsigned char *)silecsData.data.SilecsHeader_hdrBlk.device[0]._user, "%s");
	
	/* Silecs date initialization */
	SILECS_set_dt(%s,%s,%s,%s,%s,%s,%s,&silecsData.data.SilecsHeader_hdrBlk.device[0]._date);
}

"""

#=========================================================================
# Example generation
#=========================================================================

instantiationExample = """/*
 * Automatically generated Addressing example
 *"""

deviceModeDeviceListExample = """
 * This example shows how to address the register %s of block %s
 * of device %s of the class %s
 * 
 *  silecsData.%s_%s.%s.%s = ....;
 */"""

deviceModeDeviceNumberExample = """
 * This example shows how to address the register %s of block %s
 * of device 0 of the class %s
 *
 *  silecsData.%s_device[0].%s.%s = ....;
 */"""

blockModeDeviceListExample = """
 * This example shows how to address the register %s of block %s
 * of device %s of the class %s
 *
 *  silecsData.%s_%s.%s.%s = ....;
 */"""

blockModeDeviceNumberExample = """
 * This example shows how to address the register %s of block %s
 * of device 0 of the class %s
 *
 *  silecsData.%s_%s.device[0].%s = ....;
 */"""

#=========================================================================
# C code generation Sub-function
#=========================================================================

# HEADER file ------------------------------------------------------------
def cHeader(baseAddress,silecsVersion):
    return header %(silecsVersion, baseAddress)

# REGISTER definition ----------------------------------------------------
def cRegister(regName, regFormat, regDim, regDim2=1, strLen=1):
	if strLen > 1: # if it's a string
		if regDim2 > 1:    # add another pair of brackets to hold the string length in case second dimension of array is set
			return stringArray %(whichRabbitFormat[regFormat], regName, regDim, regDim2, strLen)
		elif regDim > 1:    # reuse the array2D syntax
			return regArray2d %(whichRabbitFormat[regFormat], regName, regDim, strLen)
		else:        # regular string register
			return regArray %(whichRabbitFormat[regFormat], regName, strLen)
	else:
		if regDim == 1 and regDim2 == 1:    # scalar
			return regScalar %(whichRabbitFormat[regFormat], regName)
		elif regDim > 1 and regDim2 == 1:    # array
			return regArray %(whichRabbitFormat[regFormat], regName, regDim)
		else:    # dim1>=1 and for whatever dim2, use double array syntax
			return regArray2d %(whichRabbitFormat[regFormat], regName, regDim, regDim2)

# BLOCK type definition (UDT) --------------------------------------------
def cBlockUDT(className, classVersion, blockName, regList, isFirst):
    if isFirst:
        return firstBlockUDT %(className, classVersion) + blockUDT %(regList, className, blockName)
    else:
        return blockUDT %(regList, className, blockName)
    

#============= DEVICE instantiation ============= 
   
# GLOBAL instantiation ----------------------------------------------

def cAllocationComment(plcProtocol):
	return allocationComment %(plcProtocol)

def cNBdeviceDefinition(className,deviceNumber):
    return NBdeviceDefinition %(className,deviceNumber)
     
# DEVICE MODE instantiation ----------------------------------------------

def cDeviceModeBlockInstantiation(className, blockName):
    return deviceModeBlockInstantiation %(className, blockName, blockName)

def cDeviceModeClass_deviceNumber(className, blockList):
    return deviceModeClass_deviceNumber %(blockList, className, className)

def cDeviceModeClass_deviceList(blockList, deviceList):
    return deviceModeClass_deviceList %(blockList, deviceList)

def cDeviceModeDataInstantiation(classList):
    return deviceMode_dataInstantiation %(classList)

# BLOCK MODE instantiation ----------------------------------------------

def cBlockModeDeviceInstantiation_deviceNumber(className, blockName):
    return blockModeDeviceInstantiation_deviceNumber %(className, blockName, className)

def cBlockModeDeviceInstantiation_deviceList(className, blockName, deviceName):
    return blockModeDeviceInstantiation_deviceList %(className, blockName, deviceName)

def cBlockModeBlockInstantiation(deviceList, className, blockName):
    return blockModeBlockInstantiation %(deviceList, className, blockName)
    
def cBlockModeDataInstantiation(classList):
	return blockMode_dataInstantiation %(classList)

# Init Function ----------------------------------------------
def cInitDeviceMode(silecsVersion,ieChecks,ieUser):
    dt = time.localtime(time.time())
    return initFunctionDeviceMode %("SILECS_"+silecsVersion,ieChecks,ieUser,dt[6],dt[5],dt[4],dt[3],dt[2],dt[1],dt[0])

def cInitBlockMode(silecsVersion,ieChecks,ieUser):
    dt = time.localtime(time.time())
    return initFunctionBlockMode %("SILECS_"+silecsVersion,ieChecks,ieUser,dt[6],dt[5],dt[4],dt[3],dt[2],dt[1],dt[0])
        
# Example function  ----------------------------------------------
def cExample(plcProtocol,withDeviceList,className, blockName, registerName, deviceLabel =''):

    if plcProtocol == 'DEVICE_MODE':
        if withDeviceList == True:
            # device mode - device list            
            return instantiationExample  + deviceModeDeviceListExample %(registerName, blockName, deviceLabel, className, className, deviceLabel, blockName, registerName)
        else:
            # device mode - device number
            return instantiationExample + deviceModeDeviceNumberExample %(registerName, blockName, className, className, blockName, registerName)
    else:
        if withDeviceList == True:
            # block mode - device list
            return instantiationExample + blockModeDeviceListExample %(registerName, blockName, deviceLabel, className, className, blockName, deviceLabel, registerName)
        else:
            # block mode - device number
            return instantiationExample + blockModeDeviceNumberExample %(registerName, blockName, className, className, blockName, registerName)
