// Copyright 2016 CERN and GSI
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <silecs-communication/interface/equipment/SilecsDevice.h>

#include <silecs-communication/interface/core/Context.h>
#include <silecs-communication/interface/core/SilecsService.h>
#include <silecs-communication/interface/core/WrapperAction.h>
#include <silecs-communication/interface/equipment/SilecsBlock.h>
#include <silecs-communication/interface/equipment/SilecsCluster.h>
#include <silecs-communication/interface/equipment/SilecsPLC.h>
#include <silecs-communication/interface/equipment/SilecsRegister.h>
#include <silecs-communication/interface/equipment/PLCRegister.h>
#include <silecs-communication/interface/utility/SilecsException.h>
#include <silecs-communication/interface/utility/SilecsLog.h>
#include <silecs-communication/interface/utility/XMLParser.h>

#include <silecs-communication/interface/equipment/CNVRegister.h>

#include <sstream>

namespace Silecs
{
	Device::Device(PLC* thePLC, std::string deviceName, std::string deviceAddress, boost::ptr_vector<ElementXML>& blockNodes) : thePLC_ (thePLC)
	{
		// Update Device ElementXML attributes
        label_ = deviceName;
		StringUtilities::fromString(address_,deviceAddress );

        LOG(ALLOC) << "Creating SilecsDevice: " << label_;

        boost::ptr_vector<ElementXML>::const_iterator blockIter;
        for(blockIter = blockNodes.begin(); blockIter != blockNodes.end(); blockIter++)
		{
			std::string blockName = blockIter->getAttribute("name");
			LOG(ALLOC) << "Adding Block to Device: " << blockName;
			AccessType accessType = Block::whichAccessType(blockIter->getAttribute("mode"));
			std::vector< boost::shared_ptr<ElementXML> > registerNodes = blockIter->childList_;
			instantiateRegisters(blockName, accessType, registerNodes);
		}
	}

	Device::~Device()
	{
		LOG(ALLOC) << "Device (delete): " << label_;

		// Remove the register instances related to this Device
		registerVectorType::iterator pRegisterIter;
		for(pRegisterIter = registerCol_.begin(); pRegisterIter != registerCol_.end(); ++pRegisterIter) delete pRegisterIter->second;
		registerCol_.clear();

		// Register instances have been deleted before, just clear the block-register Map.
	    blockRegisterMap_.clear();
	}


	PLC* Device::getPLC()
	{ return thePLC_;
	}


	Register* Device::getRegister(std::string registerName)
	{
		registerVectorType::iterator iter;
		for(iter = registerCol_.begin(); iter != registerCol_.end(); ++iter)
		{
			if (iter->first == registerName)
			{
				return iter->second;
			}
		}
        throw SilecsException(__FILE__, __LINE__, PARAM_UNKNOWN_REGISTER_NAME, registerName);
        return NULL; // to suppress warning
	}


	std::vector<Register*>& Device::getRegisterCollection(std::string blockName)
	{
		blockRegisterMapType::iterator iter = blockRegisterMap_.find(blockName);
        if (iter == blockRegisterMap_.end())
        { throw SilecsException(__FILE__, __LINE__, PARAM_UNKNOWN_BLOCK_NAME, blockName);
        }
        return iter->second;
	}


	int Device::recv(std::string blockName)
	{
		//Synchronous device data receive

		//Context of the transaction:
		//this = the device to be treated,
		//false = is not for registers synchronization (normal transaction)
		Context context(this, false);

		//Execute the receive action for the required device (from the current thread)
		return getBlock(blockName, Input)->getTask()->execute(&context);
	}


	int Device::send(std::string blockName)
	{
		//Synchronous device data send

		//Context of the transaction:
		//this = the device to be treated,
		//false = is not for registers synchronization (normal transaction)
		Context context(this, false);

		//Execute the receive action for the required device (from the current thread)
		return getBlock(blockName, Output)->getTask()->execute(&context);
	}


	std::string Device::getLabel() { return label_; }


	std::string Device::getRegisterList()
	{
		std::string registerList = "";
		registerVectorType::iterator pRegisterIter;
		for(pRegisterIter = registerCol_.begin(); pRegisterIter != registerCol_.end(); ++pRegisterIter)
		{	if (!registerList.empty()) registerList.append(" ");
			registerList.append(pRegisterIter->second->getName());
		}
		return registerList;
	}


	Register* Device::instantiateRegister(const boost::shared_ptr<ElementXML>& registerNode)
	{
		switch (getPLC()->getBrandID())
		{
			// SIEMENS PLCs use Motorola memory convention (BigEndianRegister)
			case Siemens:
				return new S7Register(this, registerNode);

			// SCHNEIDER PLCs use Intel memory convention (LittleEndianRegister)
			// RABBIT microcontrollers use Intel memory convention (LittleEndianRegister)
			case Schneider:
			case Digi:
				return new UnityRegister(this, registerNode);

			// BECKHOFF PLCs use different memory convention depending on PLC type:
			// . CX model: same convention than Unity
			// . BC model: special convention for 64 bits data (float, ..): swap only the two 16bits words together (no bytes swap).
			case Beckhoff:
				return new TwinCATRegister(this, registerNode);

#ifdef NI_SUPPORT_ENABLED
			//NI Shared variables use specific convention
            case Ni:
                return new CNVRegister(this, registerNode);
#endif

			default:
				throw SilecsException(__FILE__, __LINE__, DATA_UNKNOWN_PLC_MANUFACTURER, getPLC()->getBrand());

		} //switch
	}


	void Device::instantiateRegisters(std::string blockName, AccessType accessType, std::vector< boost::shared_ptr<ElementXML> >& registerNodes)
	{
		std::vector<Register*> registerCol;

		//LOG(SETUP) << "Ordering registers for block: " <<  blockName << " on Device: " << getLabel();

		std::vector< boost::shared_ptr<ElementXML> >::const_iterator registerIter;
		for(registerIter = registerNodes.begin(); registerIter != registerNodes.end(); registerIter++)
		{
			std::string registerName = (*registerIter)->getAttribute("name");
			Register* pReg = instantiateRegister(*registerIter);
			pReg->setAccessType(accessType);
			pReg->setBlockName(blockName);
			registerCol.push_back(pReg);
			registerCol_.push_back(std::make_pair(registerName, pReg));
		}
		blockRegisterMap_.insert(std::make_pair(blockName, registerCol));
	}

	Block* Device::getBlock(const std::string blockName, AccessType accessType)
	{
		if( ! hasBlock(blockName) )
		{
			std::ostringstream errorMessage;
			errorMessage << "Block not found! The block '" << blockName << "' does not exist on the device '" << label_ <<  "'.";
			throw SilecsException(__FILE__, __LINE__,errorMessage.str());
		}
		return thePLC_->getBlock(blockName, accessType);
	}

	bool Device::hasBlock(const std::string blockName)
	{
		blockRegisterMapType::iterator iter;
		iter = blockRegisterMap_.find(blockName);
		return ( iter != blockRegisterMap_.end() );
	}


	void Device::importRegisters(Block* pBlock, void* pBuffer, timeval ts, Context* pContext)
	{
		std::vector<Register*>& registerCol = getRegisterCollection(pBlock->getName());

		if (pContext->isForSynchronization())
		{	// Transaction has been done within the context of retentive registers synchronization,
			// do not update volatile registers (client has to force a transaction himself if required).
			for (unsigned int i=0; i<registerCol.size(); i++)
			{	
				Register* pReg = registerCol[i];
				if (pReg->isRetentive())
				{
					if(pReg->getFormat() == String)
				        pReg->importString(pBuffer, ts);
				    else // all registers but those having string format
				        pReg->importValue(pBuffer, ts);
				}
			}
		}
		else
		{	// It's a normal transaction from the client process,
			// all registers of the block have to be updated.
			for (unsigned int i=0; i<registerCol.size(); i++)
			{
				if(registerCol[i]->getFormat() == String)
                    registerCol[i]->importString(pBuffer, ts);
                else // all registers but those having string format
                    registerCol[i]->importValue(pBuffer, ts);
			}
		}
	}

	void Device::exportRegisters(Block* pBlock, void* pBuffer, Context* pContext)
	{
		std::vector<Register*>& registerCol = getRegisterCollection(pBlock->getName());

		if (pContext->isForSynchronization())
		{	// Transaction has been done within the context of retentive registers synchronization,
			// do not update volatile register (client has to force a transaction himself if required).
			for (unsigned int i=0; i<registerCol.size(); i++)
			{
			    Register* pReg = registerCol[i];
			    if (pReg->isRetentive())
			    {
                    if(pReg->getFormat() == String)
                        pReg->exportString(pBuffer);
                    else    // all registers but those having string format
                        pReg->exportValue(pBuffer);
			    }
			}
		}
		else
		{	// It's a normal transaction from the client process,
			// all registers of the block have to be updated.
			for (unsigned int i=0; i<registerCol.size(); i++)
			{
			    if(registerCol[i]->getFormat() == String)
			        registerCol[i]->exportString(pBuffer);
			    else
			        registerCol[i]->exportValue(pBuffer);
			}
		}
	}

	void Device::copyInToOut(const std::string blockName)
	{
		std::vector<Register*>& registerCol = getRegisterCollection(blockName);
		for (unsigned int i=0; i<registerCol.size(); i++)
		{	Register* pReg = registerCol[i];
			if (pReg->hasInputAccess() && pReg->hasOutputAccess())
				pReg->copyValue();
		}

	}

} // namespace
